#  Copyright 2019, 1533 Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

import asyncio

import cbor
import logging
import websockets
from concurrent.futures import ThreadPoolExecutor

from .consumers import *
from .exceptions import ClientException
from .id_factory import StreamIdFactory
from .opts import *
from .record_data import *
from .sync import *

log = logging.getLogger('driveline')

MAX_ALIASES = 256


async def retry(fn_coro_or_future, max_retries=0, retry_duration=0.1, exceptions=None):
    exceptions = exceptions or [Exception]
    if max_retries <= 0:
        max_retries = 0x10000000
    retry_count = 0
    while retry_count <= max_retries:
        retry_count += 1
        try:
            return await fn_coro_or_future()
        except tuple(exceptions) as e:
            if retry_count == max_retries:
                raise e
            log.error('caught exception, retrying...')
        time_wait = retry_duration * (2 ^ retry_count - 1) / 2
        log.error('retry[%s]= %ss', retry_count, time_wait)
        await asyncio.sleep(time_wait)


class Stream(AutoSync):
    __slots__ = ['client', 'stream_id']

    def __init__(self, client, stream_id):
        super().__init__()
        self.client = client
        self.stream_id = stream_id

    @synchronous()
    async def append(self, record: any, **options):
        return await self.client.append(self.stream_id, record, **options)

    @synchronous()
    async def close(self):
        return await self.client.close_stream(self)

    @synchronous()
    async def truncate(self):
        return await self.client.truncate(self.stream_id)

    def get_loop(self):
        return self.client.get_loop()


# noinspection PyNoneFunctionAssignment
class DrivelineClient(AutoSync):
    __slots__ = ['consumer_id', 'consumers', 'endpoint', 'is_running', 'loop', 'stream_id_factory', 'ws', 'io_loop']

    def __init__(self, endpoint: str, loop=None):
        super().__init__()
        if loop is None:
            self.loop = asyncio.get_event_loop()
        self.endpoint = endpoint
        self.is_running = False
        self.stream_id_factory = StreamIdFactory(MAX_ALIASES)
        self.consumer_id = 0
        self.consumers = dict()
        self.io_loop = None
        self.ws = None

    @synchronous()
    async def continuous_query(self, dql: str, handler,
                               executor: ThreadPoolExecutor = None, max_concurrency: int = 1, **options) -> Query:

        consumer = Query(self, self._next_consumer_id(), dql, True, handler,
                         executor=executor, max_concurrency=max_concurrency, **options)
        self._register_consumer(consumer)
        await self._query(consumer)
        return consumer.result()

    @synchronous()
    async def query(self, dql: str, handler,
                    executor: ThreadPoolExecutor = None, max_concurrency: int = 1, **options) -> Query:
        consumer = Query(self, self._next_consumer_id(), dql, False, handler,
                         executor=executor, max_concurrency=max_concurrency, **options)
        self._register_consumer(consumer)
        await self._query(consumer)
        return consumer.result()

    async def _query(self, consumer: Query):
        if consumer.is_continuous:
            cmd = 'sq'
        else:
            cmd = 'qq'
        await self._send(cmd, consumer.consumer_id, self._encode_query_options(**consumer.options), consumer.dql)

    def _encode_query_options(self, from_record_id=None, **_options):
        opts = []
        if from_record_id is not None:
            opts.append(TAG_READ_ID)
            opts.append(from_record_id)
        if len(opts) > 0:
            return opts
        else:
            return None

    @synchronous()
    async def cancel(self, query: Query, **_options):
        assert query is not None
        exists = self._unregister_consumer(query)
        if not exists:
            return
        await self._send('can', query.consumer_id, None)

    @synchronous()
    async def append(self, stream: str, record: any, **options):
        assert stream is not None
        await self._send('app', stream, self._encode_append_options(**options), cbor.dumps(record))

    def _encode_append_options(self, **_options):
        return None

    @synchronous()
    async def truncate(self, stream):
        assert stream is not None
        await self._send('trc', None, stream)

    @synchronous()
    async def store(self, key: str, record: any, **options):
        assert key is not None
        await self._send('st', key, self._encode_store_options(**options), cbor.dumps(record))

    def _encode_store_options(self, ttl=None, cas_id=None, **_options):
        opts = []
        if cas_id is not None:
            opts.append(TAG_STORE_CAS_ID)
            opts.append(cas_id)
        if ttl is not None:
            opts.append(TAG_STORE_TTL)
            opts.append(int(ttl))
        if len(opts) > 0:
            return opts
        else:
            return None

    @synchronous()
    async def load(self, key: str, **options):
        consumer = LoadConsumer(self, self._next_consumer_id(), key, **options)
        self._register_consumer(consumer)
        await self._load(consumer)
        return consumer.result()

    async def _load(self, consumer: LoadConsumer):
        await self._send('ld', consumer.consumer_id, self._encode_load_options(**consumer.options), consumer.key)

    def _encode_load_options(self, **_options):
        return None

    async def list_streams(self, stream_pattern: str, **options):
        consumer = ListConsumer(self, self._next_consumer_id(), stream_pattern, **options)
        self._register_consumer(consumer)
        await self._list('sls', consumer)
        return consumer.result()

    async def list_keys(self, keys_pattern: str, **options):
        consumer = ListConsumer(self, self._next_consumer_id(), keys_pattern, **options)
        self._register_consumer(consumer)
        await self._list('lst', consumer)
        return consumer.result()

    async def _list(self, command: str, consumer: ListConsumer):
        await self._send(command, consumer.consumer_id, self._encode_list_options(**consumer.options), consumer.pattern)

    def _encode_list_options(self, **_options):
        return None

    @synchronous()
    async def remove(self, key: str, **options):
        await self._send('rm', self._encode_remove_options(**options), key)

    @synchronous()
    async def remove_matches(self, key_pattern: str, **options):
        await self._send('rmk', self._encode_remove_options(**options), key_pattern)

    def _encode_remove_options(self, **_options):
        return None

    @synchronous()
    async def sync(self):
        consumer = SyncConsumer(self, self._next_consumer_id())
        self._register_consumer(consumer)
        await self._sync(consumer)
        return consumer.result()

    async def _sync(self, consumer: SyncConsumer):
        await self._send('syn', consumer.consumer_id)

    async def _define(self, stream_id, stream_name):
        return await self._send('def', stream_id, stream_name)

    @synchronous()
    async def open_stream(self, stream_name: str):
        assert stream_name is not None
        stream_id, is_alias = self.stream_id_factory.next(stream_name)
        if is_alias:
            await self._define(stream_id, stream_name)
        return Stream(self, stream_id)

    @synchronous()
    def close_stream(self, stream: Stream):
        assert stream is not None
        self.stream_id_factory.release(stream.id)

    @synchronous()
    async def close(self):
        self.is_running = False
        if not self.stopped.done():
            self.stopped.set_result(True)

    def get_loop(self):
        return self.loop

    def _next_consumer_id(self):
        consumer_id = self.consumer_id
        self.consumer_id += 1
        return consumer_id

    def _register_consumer(self, consumer: Consumer):
        self.consumers[consumer.consumer_id] = consumer

    def _unregister_consumer(self, consumer: Consumer):
        try:
            del self.consumers[consumer.consumer_id]
            return True
        except KeyboardInterrupt:
            raise
        except KeyError:
            return False

    async def _connection_did_close(self):
        log.debug('connection did close')
        for consumer in list(self.consumers.values()):
            is_ok = await consumer.on_disconnect()
            if not is_ok:
                self._unregister_consumer(consumer)

    async def _connection_did_connect(self):
        log.debug('connection did connect')
        for stream_id, stream_name in self.stream_id_factory.aliases.items():
            await self._define(stream_id, stream_name)

        for consumer in list(self.consumers.values()):
            is_ok = await consumer.on_reconnect()
            if not is_ok:
                self._unregister_consumer(consumer)

    async def _message_receiver(self, ws):
        log.debug('receiving messages')
        while self.is_running:
            try:
                msg = await asyncio.wait_for(ws.recv(), timeout=4)
            except websockets.exceptions.ConnectionClosed:
                log.error('connection closed')
                return
            except asyncio.streams.IncompleteReadError:
                # connection failed
                log.exception("failed to read messages")
                return
            except asyncio.TimeoutError:
                log.debug('pinging server')
                # No data in 1 second, check the connection.
                try:
                    pong_waiter = await ws.ping()
                    await asyncio.wait_for(pong_waiter, timeout=1)
                except asyncio.TimeoutError:
                    self.ws = asyncio.Future()
                    raise ConnectionAbortedError('ping failed')
            else:
                try:
                    await self._handle_message(msg)
                except KeyboardInterrupt:
                    raise
                except Exception:
                    log.exception('error processing message')

    async def _handle_message(self, raw_message: bytes):
        msg = cbor.loads(raw_message)
        if len(msg) <= 1:
            log.warning('malformed server message')
            return
        cmd = msg[0]
        if cmd == 'data':
            _, consumer_id, tags, *records = msg
            return await self._dispatch_message(consumer_id, tags=tags, records=records)
        elif cmd == 'err':
            _, consumer_id, error = msg
            return await self._dispatch_message(consumer_id, error=error)
        elif cmd == 'syn':
            _, consumer_id = msg
            return await self._dispatch_message(consumer_id, tags=None, records=[])
        else:
            log.warning('unknown server message')

    async def _dispatch_message(self, consumer_id, **message):
        consumer: Consumer = self.consumers.get(consumer_id)
        if consumer is None:
            log.warning("received data for non-existing consumer id=%s", consumer_id)
            return

        try:
            keep_running = await consumer.on_message(**message)
        except KeyboardInterrupt:
            raise
        except Exception:
            log.exception('error processing message; stopping consumer')
            keep_running = False

        if not keep_running:
            self._unregister_consumer(consumer)

    async def _send(self, *args):
        log.debug('sending message %s', args)
        data = cbor.dumps(args)
        ws = await self.ws
        await ws.send(data)

    async def _run_io_loop(self, started: asyncio.Future):
        connection_opts = dict(
            subprotocols=('driveline',),
            compression=None,
            # read_limit sets the high-water limit of the buffer for incoming bytes
            read_limit=2 ** 10,
            # write_limit sets the high-water limit of the buffer for outgoing bytes
            write_limit=2 ** 10,
            # max_size enforces the maximum size for incoming messages in bytes
            max_size=None,  # 2 ** 20,
            # max_queue sets the maximum length of the queue that holds incoming messages
            max_queue=10,
            # timeout  defines the maximum wait time in seconds for completing the closing handshake
            timeout=4,
        )
        while self.is_running:
            self.ws = asyncio.Future()
            log.debug('connecting')
            ws = await retry(
                lambda: websockets.connect(self.endpoint, **connection_opts),
                retry_duration=0.75  # second
            )
            log.debug('connected')
            self.ws.set_result(ws)
            receiver = asyncio.ensure_future(self._message_receiver(ws), loop=self.loop)
            if not started.done():
                started.set_result(True)
            await self._connection_did_connect()
            try:
                await asyncio.wait([receiver], loop=self.loop)
            except (websockets.ConnectionClosed, ConnectionAbortedError):
                log.debug('connection lost')
                pass
            self.ws = asyncio.Future()
            await asyncio.ensure_future(self._close_ws(ws), loop=self.loop)
            log.debug('closed connection')
            await self._connection_did_close()

    async def _close_ws(self, ws):
        try:
            await ws.close()
        except:
            log.exception('error while closing connection')

    async def __aenter__(self):
        self.is_running = True
        started = asyncio.Future()
        self.stopped = asyncio.Future()
        self.io_loop = asyncio.ensure_future(self._run_io_loop(started))
        await started
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        log.debug('awaiting completion')
        try:
            await asyncio.wait([self.stopped, self.io_loop], return_when=asyncio.FIRST_COMPLETED)
        except KeyboardInterrupt:
            raise
        log.debug('received completion')


def run_async(coro_or_future):
    try:
        if hasattr(asyncio, 'run'):
            return asyncio.run(coro_or_future)
        else:
            loop = asyncio.get_event_loop()
            return loop.run_until_complete(coro_or_future)
    except KeyboardInterrupt:
        pass


__all__ = [
    'DrivelineClient',
    'ClientException',
    'ListResult',
    'Record',
    'Stream',
    'Query',
    'run_async',
]



#  Copyright 2019, 1533 Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

import asyncio


class SyncProxy(object):

    def __init__(self, async_target):
        self.async_target = async_target

    def __getattr__(self, item):
        method = getattr(self.async_target, item + '_sync', None)
        if method is not None:
            return method
        return getattr(self.async_target, item)


class AutoSync(object):

    def __getattr__(self, item):
        if not item.endswith('_sync'):
            raise AttributeError(item)

        method = getattr(self, item[:-5], None)
        config = getattr(method, '__auto_sync__', None)
        if config is None:
            raise AttributeError(item)

        sync_method = self._synchronized(method, **config)
        setattr(self, item, sync_method)
        return sync_method

    def _synchronized(self, fn, **_config):
        def wrapper(*args, **kwds):
            return asyncio.run_coroutine_threadsafe(fn(*args, **kwds), loop=self.get_loop()).result()

        return wrapper

    def sync_proxy(self):
        return SyncProxy(self)


def synchronous(**config):
    def decorator(func):
        func.__auto_sync__ = config
        return func

    return decorator


__all__ = [
    'AutoSync',
    'SyncProxy',
    'synchronous',
]



import logging

from driveline import DrivelineClient, run_async
from settings import WS_ENDPOINT

logging.basicConfig(level=logging.INFO)
log = logging.getLogger('example-3')

KEY_NAME = 'key/%d'


async def main():
    async with DrivelineClient(WS_ENDPOINT) as client:
        log.info('***** starting *****')
        for i in range(1, 100):
            key_name = KEY_NAME % i
            log.info('storing document to key %s', key_name)
            document = dict(index=i, keyname=key_name)
            await client.store(key_name, document)
        log.info('***** done *****')
        await client.close()


if __name__ == '__main__':
    run_async(main())



import logging

from driveline import DrivelineClient, run_async
from settings import WS_ENDPOINT

logging.basicConfig(level=logging.INFO)
log = logging.getLogger('example-4')


async def main():
    async with DrivelineClient(WS_ENDPOINT) as client:
        log.info('***** started *****')
        await show_list(client, 'streams')
        await show_list(client, 'keys')
        log.info('***** done *****')
        await client.close()


async def show_list(client, kind):
    log.info('listing %s in driveline:', kind)
    count = 0
    async for entry_name in await getattr(client, 'list_%s' % kind)('*'):
        log.info('found %s "%s"', kind[:-1], entry_name)
        count += 1
    if count == 0:
        log.info('no %s could be found', kind)
    else:
        log.info('found %d %s', count, kind)

    log.info('done listing %s', kind)


if __name__ == '__main__':
    run_async(main())



#  Copyright 2019, 1533 Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

from .exceptions import ClientException
import cbor


class Record(object):
    @property
    def record(self):
        raise NotImplemented()

    @property
    def raw_record(self):
        raise NotImplemented()

    @property
    def record_id(self):
        raise NotImplemented()


class DataRecord(Record):
    __slots__ = ['_record', '_record_id']

    def __init__(self, record, record_id):
        self._record = record
        self._record_id = record_id

    @property
    def record(self):
        if self._record is None:
            return None
        return cbor.loads(self._record)

    @property
    def raw_record(self):
        return self._record

    @property
    def record_id(self):
        return self._record_id


class ErrorRecord(Record):
    __slots__ = ['error']

    def __init__(self, error):
        self.error = error

    @property
    def record(self):
        raise ClientException(self.error)

    @property
    def raw_record(self):
        raise ClientException(self.error)

    @property
    def record_id(self):
        raise ClientException(self.error)


__all__ = [
    'DataRecord',
    'ErrorRecord',
    'Record',
]



#  Copyright 2019, 1533 Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

class StreamIdFactory(object):
    __slots__ = ['aliases', 'count', 'free_aliases', 'stream_ids']

    def __init__(self, count: int):
        assert count > 0
        self.stream_ids = list(range(count, -1, -1))
        self.count = count
        self.free_aliases = count
        self.aliases = dict()

    def next(self, stream_name: str):
        if self.free_aliases == 0:
            return stream_name, False
        stream_id = self.stream_ids[self.free_aliases]
        self.free_aliases -= 1
        self.aliases[stream_id] = stream_name
        return stream_id, True

    def release(self, stream_id):
        if isinstance(stream_id, str):
            return
        del self.aliases[stream_id]
        self.free_aliases += 1
        self.stream_ids[self.free_aliases] = stream_id


__all__ = [
    'StreamIdFactory',
]



import asyncio
import logging

from driveline import DrivelineClient, run_async
from settings import WS_ENDPOINT

logging.basicConfig(level=logging.INFO)
log = logging.getLogger('example-5')

TEST_TTL = 100  # in millisecond
WAIT_DURATION = TEST_TTL + 50  # in millisecond


async def main():
    """ this simple tests demonstrates the use of TTL with the key-value store
    """
    async with DrivelineClient(WS_ENDPOINT) as client:
        log.info('storing document')
        await client.store('ephemeral', dict(test='me'), ttl=TEST_TTL)

        log.info('waiting for document expiry')
        await asyncio.sleep(WAIT_DURATION / 1000.0)

        log.info('fetching document')
        future_result = await client.load('ephemeral')
        # await the future
        result = await future_result
        if result.record is not None:
            log.error('document is %s. After TTL expiry it should be None', result.record)
        else:
            log.info('document is None, which respects the TTL configuration')
        await client.close()


if __name__ == '__main__':
    run_async(main())



import logging

from driveline import DrivelineClient, run_async
from driveline.video import VideoHandler
from settings import WS_ENDPOINT

logging.basicConfig(level=logging.INFO)
log = logging.getLogger('example-6')


def message_cb(frame):
    log.info('received frame: %s', frame)


async def main():
    """ example demonstrating the use of driveline-video in conjunction with driveline
    """
    async with DrivelineClient(WS_ENDPOINT) as client:
        await client.continuous_query('SELECT * FROM STREAM video_stream', VideoHandler(message_cb))


if __name__ == '__main__':
    run_async(main())



import logging

from driveline import DrivelineClient, run_async
from settings import WS_ENDPOINT

logging.basicConfig(level=logging.INFO)
log = logging.getLogger('example-2')


def query_cb(result):
    try:
        log.info("received document: %s", result.record)
    except Exception as e:
        log.error('subscription error: %s', e)


async def main():
    """ this example demonstrates how to set a subscriber.
        you should run it alongside example-1
    """
    async with DrivelineClient(WS_ENDPOINT) as client:
        log.debug('***** start *****')
        await client.continuous_query('SELECT * FROM STREAM "hello-world-stream"', query_cb)


if __name__ == '__main__':
    run_async(main())



import asyncio

import logging
import random

from driveline import DrivelineClient, run_async
from settings import WS_ENDPOINT

logging.basicConfig(level=logging.INFO)
log = logging.getLogger('example-1')


async def main():
    """ this example publishes every second a record of the form:
        { index: number of second since start, msg: message }
    """
    log.info('***** start *****')
    nonce = ''.join([chr(random.randint(ord('a'), ord('z'))) for _ in range(0, 16)])
    async with DrivelineClient(WS_ENDPOINT) as client:
        demo = await client.open_stream('hello-world-stream')
        i = 0
        while True:
            document = dict(index=i, msg='hello world at %s' % i, nonce=nonce)
            await demo.append(document)
            log.info('published object index %d to hello-world-stream', i)
            i += 1
            await asyncio.sleep(1)


if __name__ == '__main__':
    run_async(main())



import pathlib
import setuptools

VERSION = '0.14.1'

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setuptools.setup(
    name='driveline',
    version=VERSION,
    author='1533 Systems',
    author_email='info@1533.io',
    description='Driveline client',
    long_description=README,
    long_description_content_type='text/markdown',
    license='MIT',
    url='https://github.com/1533-systems/python3-sdk',
    packages=setuptools.find_packages(),
    keywords="driveline performance database streaming document",
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6.0',
    install_requires=[
        'websockets~=7.0',
        'cbor~=1.0',
    ]
)



#  Copyright 2019, 1533 Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

import asyncio

import cbor
import logging
from concurrent.futures import ThreadPoolExecutor

from .exceptions import ClientException
from .opts import *
from .record_data import *

log = logging.getLogger('driveline')


class Consumer(object):
    __slots__ = ['_client', 'consumer_id', 'options']

    def __init__(self, client, consumer_id):
        self._client = client
        self.consumer_id = consumer_id
        self.options = {}

    @property
    def client(self):
        return self._client

    async def on_reconnect(self):
        return False

    async def on_disconnect(self):
        return False

    def _build_data_records(self, records, tags):
        log.debug('processing tags: %s for %d record (s)', tags, len(records))
        tags = iter(tags or [])
        record_ids = []
        try:
            while True:
                key = next(tags)
                value = next(tags)
                if key == TAG_ID:
                    record_ids = value
        except StopIteration:
            pass
        if len(record_ids) == 0:
            record_ids = [None for _ in range(len(records))]
        else:
            log.debug('got record_ids: %s', record_ids)

        for record, record_id in zip(records, record_ids):
            yield DataRecord(record, record_id)

    async def on_message(self, records=None, tags=None, error=None):
        raise NotImplemented()


class LoadConsumer(Consumer):
    __slots__ = ['key', 'promise']

    def __init__(self, client, consumer_id, key, **options):
        super().__init__(client, consumer_id)
        self._set_read_options(**options)
        self.key = key
        self.promise = asyncio.Future()

    def _set_read_options(self, from_record_id=None):
        self.options.update(from_record_id=from_record_id)

    def result(self):
        return self.promise

    async def on_disconnect(self):
        return True

    async def on_reconnect(self):
        await self.client._load(self)
        return True

    async def on_message(self, records=None, tags=None, error=None):
        log.debug('received load response w/ %d record(s)', len(records))
        if error is not None:
            log.error('something went wrong, resolving error')
            self.promise.set_exception(ClientException(error))
            return False
        for payload in self._build_data_records(records, tags):
            log.debug('resolving the future with %s', payload)
            self.promise.set_result(payload)
            break
        log.debug('resolved future')
        return False


class ListResult(object):
    __slots__ = ['queue']

    def __init__(self, queue: asyncio.Queue):
        self.queue = queue

    def __aiter__(self):
        return self

    async def __anext__(self):
        name, error = await self.queue.get()
        if error is not None:
            raise error
        if name is None:
            raise StopAsyncIteration
        return name


class ListConsumer(Consumer):
    __slots__ = ['pattern', 'queue']

    def __init__(self, client, consumer_id, pattern, **options):
        super().__init__(client, consumer_id)
        assert pattern is not None and len(pattern) > 0
        self.pattern = pattern
        self.queue = asyncio.Queue()
        self._set_list_options(**options)

    def _set_list_options(self, **options):
        if len(options) > 0:
            log.warning('ignoring list options %s', options)

    def result(self):
        return ListResult(self.queue)

    async def on_disconnect(self):
        await self.queue.put((None, ClientException('connection lost')))
        return False

    async def on_message(self, records=None, tags=None, error=None):
        log.debug('received list data')
        if error is not None:
            await self.queue.put((None, ClientException(error)))
            return False

        records = cbor.loads(records[0])
        if len(records) == 0:
            log.debug('list data is empty. done receiving results')
            await self.queue.put((None, None))
            return False

        for m in records:
            await self.queue.put((m, None))

        return True


class Query(Consumer):
    __slots__ = ['dql', 'executor', 'handler', 'last_record_id', 'is_continuous', 'options']

    def __init__(self, client, consumer_id: int, dql: str, is_continuous: bool, handler,
                 executor=None,
                 max_concurrency=None,
                 **options):
        super().__init__(client, consumer_id)
        self._set_query_options(**options)
        assert len(dql) > 10 and handler is not None and max_concurrency > 0
        self.dql = dql
        self.is_continuous = is_continuous
        self.handler = handler
        self.executor = executor
        if self.executor is not None:
            self.executor = ThreadPoolExecutor(max_workers=int(max_concurrency))
        self.last_record_id = None

    def _set_query_options(self, from_record_id=None):
        self.options.update(from_record_id=from_record_id)

    @property
    def client(self):
        return self._client.sync_proxy()

    async def on_disconnect(self):
        return self.is_continuous

    async def on_reconnect(self):
        self.options.update(from_record_id=self.last_record_id)
        await self.client._query(self)
        return True

    def result(self):
        return self

    async def on_message(self, records=None, tags=None, error=None):
        if error is not None:
            await self.client.loop.run_in_executor(self.executor, self.handler, ErrorRecord(error))
            return False

        if not self.is_continuous and len(records) == 0:
            # stop condition for quick queries
            try:
                self.handler(None)
            except KeyboardInterrupt:
                raise
            except Exception:
                log.exception('error processing message')
            return False

        log.debug('received %d messages for query %s', len(records), self.dql)
        for m in self._build_data_records(records, tags):
            if (m.record_id or b'') < (self.last_record_id or b''):
                continue
            try:
                await self.client.loop.run_in_executor(self.executor, self.handler, m)
                self.last_record_id = m.record_id
            except KeyboardInterrupt:
                raise
            except Exception:
                log.exception('error processing message')
        return True


class SyncConsumer(Consumer):
    __slots__ = ['promise']

    def __init__(self, client, consumer_id):
        super().__init__(client, consumer_id)
        self.promise = asyncio.Future()

    def result(self):
        return self.promise

    async def on_disconnect(self):
        self.promise.set_exception(ClientException('connection lost'))
        return False

    async def on_reconnect(self):
        return False

    async def on_message(self, records=None, tags=None, error=None):
        log.debug('received load response w/ %d record(s)', len(records))
        if error is not None:
            self.promise.set_exception(ClientException(error))
            return False
        self.promise.set_result(None)
        return False


__all__ = [
    'Query',
    'Consumer',
    'LoadConsumer',
    'ListConsumer',
    'SyncConsumer',
    'ListResult',
]



#  Copyright 2019, 1533 Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

try:
    from _driveline_video import VideoDecoder
except ImportError:
    raise ImportError('driveline-video is not installed')


class VideoHandler(object):

    def __init__(self, frame_handler=None):
        self.decoder = VideoDecoder()
        self.frame_handler = frame_handler

    def __call__(self, record, *args, **kwargs):
        try:
            self.decoder.decode(record.raw_record)
            while True:
                frame = self.decoder.next_frame()
                if frame is None:
                    break
                self.on_frame(frame)
        except:
            pass

    def on_frame(self, frame):
        self.frame_handler(frame)


__all__ = [
    'VideoHandler'
]
